/*****************************************************************************
 *                        Copyright (c) 2001 Daniel Selman
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 ****************************************************************************/

package org.j3d.renderer.aviatrix3d.geom.hanim;

// External imports
import java.util.ArrayList;
import java.nio.FloatBuffer;

import javax.vecmath.Matrix4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.geom.hanim.HAnimJoint;
import org.j3d.geom.hanim.HAnimObject;

/**
 * Implementation of the joint object that does in-place software evaluation of
 * the skin mesh updates.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.1 $
 */
class SoftwareJoint extends AVJoint
{
    /** The array of flags to indicate the managed coord index */
    private boolean[] dirtyCoordinates;

    /** Have we been marked as dirty this frame? */
    private boolean dirty;

    /**
     * Create a new, default instance of the site.
     */
    SoftwareJoint()
    {
        dirty = false;
    }

    //----------------------------------------------------------
    // Methods defined by HAnimJoint
    //----------------------------------------------------------

    /**
     * Replace the existing children with the new set of children.
     *
     * @param kids The collection of child objects to now use
     * @param numValid The number kids to copy from the given array
     */
    public void setChildren(HAnimObject[] kids, int numValid)
    {
        super.setChildren(kids, numValid);

        for(int i = 0; i < numChildren; i++)
        {
            if(children[i] instanceof HAnimJoint)
                ((SoftwareJoint)children[i]).setDirtyList(dirtyCoordinates);
        }
    }

    /**
     * Add a child node to the existing collection. Duplicates and null values
     * are allowed.
     *
     * @param kid The new child instance to add
     */
    public void addChild(HAnimObject kid)
    {
        super.addChild(kid);

        if(kid instanceof HAnimJoint)
            ((SoftwareJoint)kid).setDirtyList(dirtyCoordinates);
    }

    /**
     * All the skeletal changes are in for this frame, so update the matrix
     * values now. This should not be callable by the general public. Derived
     * classes may override this method, but should call it as well to ensure
     * the internal matrices are correctly updated.
     *
     * @param parentTransform The transformation into global coordinates of
     *   the parent of this joint
     * @param parentChanged Flag to indicate that the parent transformation
     *   matrix has changed or is still the same as last call
     */
    protected void updateSkeleton(Matrix4f parentTransform,
                                  boolean parentChanged)
    {
        super.updateSkeleton(parentTransform, parentChanged);

        // now update the weighted vertices using the global matrix
        if(outputCoords == null)
            return;

        // if either of the normal items are dodgy, ignore
        if(numSourceNormals == 0)
        {
            FloatBuffer c_buf = (FloatBuffer)outputCoords;

            for(int i = 0; i < numSkinCoord; i++)
            {
                int index = skinCoordIndex[i];
                if(!dirtyCoordinates[index])
                    continue;

                float x = sourceCoords[index * 3];
                float y = sourceCoords[index * 3 + 1];
                float z = sourceCoords[index * 3 + 2];

                float curr_x = c_buf.get(index * 3);
                float curr_y = c_buf.get(index * 3 + 1);
                float curr_z = c_buf.get(index * 3 + 2);

                float out_x = globalMatrix.m00 * x + globalMatrix.m01 * y +
                              globalMatrix.m02 * z + globalMatrix.m03;
                float out_y = globalMatrix.m10 * x + globalMatrix.m11 * y +
                              globalMatrix.m12 * z + globalMatrix.m13;
                float out_z = globalMatrix.m20 * x + globalMatrix.m21 * y +
                              globalMatrix.m22 * z + globalMatrix.m23;

                c_buf.put(index * 3, curr_x + out_x * skinCoordWeight[i]);
                c_buf.put(index * 3 + 1, curr_y + out_y * skinCoordWeight[i]);
                c_buf.put(index * 3 + 2, curr_z + out_z * skinCoordWeight[i]);

                curr_x = c_buf.get(index * 3);
                curr_y = c_buf.get(index * 3 + 1);
                curr_z = c_buf.get(index * 3 + 2);
            }
        }
        else
        {
            FloatBuffer c_buf = (FloatBuffer)outputCoords;
            FloatBuffer n_buf = (FloatBuffer)outputNormals;

            for(int i = 0; i < numSkinCoord; i++)
            {
                int index = skinCoordIndex[i];
                if(!dirtyCoordinates[index])
                    continue;

                float cx = sourceCoords[index * 3];
                float cy = sourceCoords[index * 3 + 1];
                float cz = sourceCoords[index * 3 + 2];

                float nx = sourceNormals[index * 3];
                float ny = sourceNormals[index * 3 + 1];
                float nz = sourceNormals[index * 3 + 2];

                float curr_cx = c_buf.get(index * 3);
                float curr_cy = c_buf.get(index * 3 + 1);
                float curr_cz = c_buf.get(index * 3 + 2);

                float curr_nx = n_buf.get(index * 3);
                float curr_ny = n_buf.get(index * 3 + 1);
                float curr_nz = n_buf.get(index * 3 + 2);

                float out_cx = globalMatrix.m00 * cx + globalMatrix.m01 * cy +
                               globalMatrix.m02 * cz + globalMatrix.m03;
                float out_cy = globalMatrix.m10 * cx + globalMatrix.m11 * cy +
                               globalMatrix.m12 * cz + globalMatrix.m13;
                float out_cz = globalMatrix.m20 * cx + globalMatrix.m21 * cy +
                               globalMatrix.m22 * cz + globalMatrix.m23;

                float out_nx = globalMatrix.m00 * nx + globalMatrix.m01 * ny +
                               globalMatrix.m02 * nz;
                float out_ny = globalMatrix.m10 * nx + globalMatrix.m11 * ny +
                               globalMatrix.m12 * nz;
                float out_nz = globalMatrix.m20 * nx + globalMatrix.m21 * ny +
                               globalMatrix.m22 * nz;

                c_buf.put(index * 3, curr_cx + out_cx * skinCoordWeight[i]);
                c_buf.put(index * 3 + 1, curr_cy + out_cy * skinCoordWeight[i]);
                c_buf.put(index * 3 + 2, curr_cz + out_cz * skinCoordWeight[i]);

                n_buf.put(index * 3, curr_nx + out_nx * skinCoordWeight[i]);
                n_buf.put(index * 3 + 1, curr_ny + out_ny * skinCoordWeight[i]);
                n_buf.put(index * 3 + 2, curr_nz + out_nz * skinCoordWeight[i]);
            }
        }

        dirty = false;
    }

    /**
     * Send an update message to the parent, if one has not already been sent.
     */
    protected void sendUpdateMsg()
    {
        if(!dirty && dirtyCoordinates != null)
        {
            for(int i = 0; i < numSkinCoord; i++)
                dirtyCoordinates[skinCoordIndex[i]] = true;

            for(int i = 0; i < numChildren; i++)
            {
                if(children[i] instanceof HAnimJoint)
                    ((SoftwareJoint)children[i]).parentDirty();
            }

            dirty = true;
        }

        super.sendUpdateMsg();
    }

    //----------------------------------------------------------
    // Local Methods
    //----------------------------------------------------------

    /**
     * Set the dirty coordinate list for this node to play with.
     *
     * @param dirtyList Array that we mark with dirty flags
     */
    void setDirtyList(boolean[] dirtyList)
    {
        dirtyCoordinates = dirtyList;

        for(int i = 0; i < numChildren; i++)
        {
            if(children[i] instanceof HAnimJoint)
                ((SoftwareJoint)children[i]).setDirtyList(dirtyList);
        }
    }

    /**
     * Notification that the parent is dirty for this frame.
     */
    void parentDirty()
    {
        if(!dirty && dirtyCoordinates != null)
        {
            for(int i = 0; i < numSkinCoord; i++)
                dirtyCoordinates[skinCoordIndex[i]] = true;

            for(int i = 0; i < numChildren; i++)
            {
                if(children[i] instanceof HAnimJoint)
                    ((SoftwareJoint)children[i]).parentDirty();
            }

            dirty = true;
        }
    }
}
